from numpy.testing import assert_almost_equal, assert_equal
from falass import dataformat, reflect
import unittest
import numpy as np


class TestReflect(unittest.TestCase):
    def test_reflect(self):
        layer1 = dataformat.SLDPro(1., 0., 0.)
        layer2 = dataformat.SLDPro(1., 5., 0.)
        sld = [[layer1, layer2]]
        data1 = dataformat.QData(0.05, 0., 0., 0.05 * 0.05)
        data2 = dataformat.QData(0.25, 0., 0., 0.05 * 0.25)
        data3 = dataformat.QData(0.50, 0., 0., 0.05 * 0.50)
        data = [data1, data2, data3]
        a = reflect.Reflect(sld, data)
        assert_almost_equal(a.sld_profile[0][0].thick, 1.)
        assert_almost_equal(a.sld_profile[0][0].real, 0.)
        assert_almost_equal(a.sld_profile[0][0].imag, 0.)
        assert_almost_equal(a.sld_profile[0][1].thick, 1.)
        assert_almost_equal(a.sld_profile[0][1].real, 5.)
        assert_almost_equal(a.sld_profile[0][1].imag, 0.)
        assert_almost_equal(a.exp_data[0].q, 0.05)
        assert_almost_equal(a.exp_data[0].dq, 0.05 * 0.05)
        assert_almost_equal(a.exp_data[1].q, 0.25)
        assert_almost_equal(a.exp_data[1].dq, 0.25 * 0.05)
        assert_almost_equal(a.exp_data[2].q, 0.50)
        assert_almost_equal(a.exp_data[2].dq, 0.50 * 0.05)

    def test_calc_ref_basic(self):
        layer1 = dataformat.SLDPro(1., 0., 0.)
        layer2 = dataformat.SLDPro(1., 5., 0.)
        sld = [[layer1, layer2]]
        data1 = dataformat.QData(0.05, 0., 0., 0.05 * 0.05)
        data2 = dataformat.QData(0.25, 0., 0., 0.05 * 0.25)
        data3 = dataformat.QData(0.50, 0., 0., 0.05 * 0.50)
        data = [data1, data2, data3]
        a = reflect.Reflect(sld, data)
        a.calc_ref()
        assert_equal(len(a.reflect), 1)
        assert_equal(len(a.reflect[0]), 3)

    def test_calc_reflect_noq(self):
        layer1 = dataformat.SLDPro(1., 0., 0.)
        layer2 = dataformat.SLDPro(1., 5., 0.)
        sld = [[layer1, layer2]]
        ddata = []
        a = reflect.Reflect(sld, ddata)
        with self.assertRaises(ValueError) as context:
            a.calc_ref()
        self.assertTrue('No q vectors have been defined -- either read a .dat file or get q vectors.' in str(context.exception))

    def test_average_reflect(self):
        data11 = dataformat.QData(0.05, 0.1, 0., 0.05 * 0.05)
        data12 = dataformat.QData(0.25, 0.05, 0., 0.05 * 0.25)
        data13 = dataformat.QData(0.50, 0.01, 0., 0.05 * 0.50)
        data1 = [data11, data12, data13]
        data21 = dataformat.QData(0.05, 0.12, 0., 0.05 * 0.05)
        data22 = dataformat.QData(0.25, 0.03, 0., 0.05 * 0.25)
        data23 = dataformat.QData(0.50, 0.015, 0., 0.05 * 0.50)
        data2 = [data21, data22, data23]
        data31 = dataformat.QData(0.05, 0.14, 0., 0.05 * 0.05)
        data32 = dataformat.QData(0.25, 0.07, 0., 0.05 * 0.25)
        data33 = dataformat.QData(0.50, 0.005, 0., 0.05 * 0.50)
        data3 = [data31, data32, data33]
        layer1 = dataformat.SLDPro(1., 0., 0.)
        layer2 = dataformat.SLDPro(1., 5., 0.)
        sld = [[layer1, layer2]]
        ddata1 = dataformat.QData(0.05, 0., 0., 0.05 * 0.05)
        ddata2 = dataformat.QData(0.25, 0., 0., 0.05 * 0.25)
        ddata3 = dataformat.QData(0.50, 0., 0., 0.05 * 0.50)
        ddata = [ddata1, ddata2, ddata3]
        a = reflect.Reflect(sld, ddata)
        a.reflect = [data1, data2, data3]
        a.average_ref()
        assert_almost_equal(a.averagereflect[0].i, 0.12)
        assert_almost_equal(a.averagereflect[1].i, 0.05)
        assert_almost_equal(a.averagereflect[2].i, 0.01)

    def test_average_reflect_noq(self):
        layer1 = dataformat.SLDPro(1., 0., 0.)
        layer2 = dataformat.SLDPro(1., 5., 0.)
        sld = [[layer1, layer2]]
        ddata = []
        a = reflect.Reflect(sld, ddata)
        with self.assertRaises(ValueError) as context:
            a.average_ref()
        self.assertTrue('No q vectors have been defined -- either read a .dat file or get q vectors.' in str(context.exception))


    def test_make_kn(self):
        layer1 = dataformat.SLDPro(5., 5., 0.)
        layer2 = dataformat.SLDPro(5., 0., 0.)
        sld_profile = [layer1, layer2]
        exp_data = [0.05, 0.25, 0.5]
        layers = np.zeros((len(sld_profile), 4))
        for i in range(0, len(sld_profile)):
            layers[i][0] = sld_profile[i].thick
            layers[i][1] = sld_profile[i].real
            layers[i][2] = sld_profile[i].imag
            layers[i][3] = 0
        qvals = np.asfarray(exp_data).ravel()
        nlayers = len(sld_profile) - 2
        npnts = qvals.size
        kn = reflect.make_kn(npnts, nlayers, layers, qvals)
        assert_almost_equal(kn, [[0.025 + 0j, 7.9266940191 + 0j], [0.125 + 0j, 7.927640133 + 0j],
                                 [0.25 + 0j, 7.93059601 + 0j]])


    def test_knext_and_rj(self):
        kn = np.array([[0.025 + 0j, 7.9266940191 + 0j], [0.125 + 0j, 7.927640133 + 0j],
                       [0.25 + 0j, 7.93059601 + 0j]])
        k = np.array([0.025 + 0j, 0.125 + 0j, 0.25 + 0j])
        idx = 1
        k_next, rj = reflect.knext_and_rj(kn, idx, k)
        assert_almost_equal(k_next, np.array([7.9266940191 + 0j,  7.927640133 + 0j, 7.93059601 + 0j]))
        assert_almost_equal(rj, np.array([-1.211500325 + 0j, -2.610174734 + 0j, -6.8181020565 + 0j]))



from numpy.testing import assert_almost_equal
from falass import compare, dataformat, reflect
import numpy as np
import unittest

class TestCompare(unittest.TestCase):
    def test_compare(self):
        data1 = dataformat.QData(0.05, 3., 0.3, 0.05 * 0.05)
        data2 = dataformat.QData(0.25, 2., 0.2, 0.05 * 0.25)
        data3 = dataformat.QData(0.50, 1., 0.1, 0.05 * 0.50)
        data = [data1, data2, data3]
        sdata1 = dataformat.QData(0.05, 1., 0.1, 0.05 * 0.05)
        sdata2 = dataformat.QData(0.25, 2., 0.2, 0.05 * 0.25)
        sdata3 = dataformat.QData(0.50, 3., 0.3, 0.05 * 0.50)
        sdata = [sdata1, sdata2, sdata3]
        a = compare.Compare(data, sdata, 1., 0.)
        assert_almost_equal(a.exp_data[0].q, 0.05)
        assert_almost_equal(a.exp_data[0].i, 3.)
        assert_almost_equal(a.exp_data[0].di, 0.3)
        assert_almost_equal(a.exp_data[0].dq, 0.05 * 0.05)
        assert_almost_equal(a.exp_data[1].q, 0.25)
        assert_almost_equal(a.exp_data[1].i, 2.)
        assert_almost_equal(a.exp_data[1].di, 0.2)
        assert_almost_equal(a.exp_data[1].dq, 0.05 * 0.25)
        assert_almost_equal(a.exp_data[2].q, 0.50)
        assert_almost_equal(a.exp_data[2].i, 1.)
        assert_almost_equal(a.exp_data[2].di, 0.1)
        assert_almost_equal(a.exp_data[2].dq, 0.05 * 0.50)
        assert_almost_equal(a.sim_data[0].q, 0.05)
        assert_almost_equal(a.sim_data[0].i, 1.)
        assert_almost_equal(a.sim_data[0].di, 0.1)
        assert_almost_equal(a.sim_data[0].dq, 0.05 * 0.05)
        assert_almost_equal(a.sim_data[1].q, 0.25)
        assert_almost_equal(a.sim_data[1].i, 2.)
        assert_almost_equal(a.sim_data[1].di, 0.2)
        assert_almost_equal(a.sim_data[1].dq, 0.05 * 0.25)
        assert_almost_equal(a.sim_data[2].q, 0.50)
        assert_almost_equal(a.sim_data[2].i, 3.)
        assert_almost_equal(a.sim_data[2].di, 0.3)
        assert_almost_equal(a.sim_data[2].dq, 0.05 * 0.50)
        assert_almost_equal(a.scale, 1.)
        assert_almost_equal(a.background, 0.)

    def test_fit_noq(self):
        data11 = dataformat.QData(0.05, 0.1, 0., 0.05 * 0.05)
        data12 = dataformat.QData(0.25, 0.05, 0., 0.05 * 0.25)
        data13 = dataformat.QData(0.50, 0.01, 0., 0.05 * 0.50)
        data1 = [data11, data12, data13]
        data21 = dataformat.QData(0.05, 0.12, 0., 0.05 * 0.05)
        data22 = dataformat.QData(0.25, 0.03, 0., 0.05 * 0.25)
        data23 = dataformat.QData(0.50, 0.015, 0., 0.05 * 0.50)
        data2 = [data21, data22, data23]
        data31 = dataformat.QData(0.05, 0.14, 0., 0.05 * 0.05)
        data32 = dataformat.QData(0.25, 0.07, 0., 0.05 * 0.25)
        data33 = dataformat.QData(0.50, 0.005, 0., 0.05 * 0.50)
        data3 = [data31, data32, data33]
        layer1 = dataformat.SLDPro(1., 0., 0.)
        layer2 = dataformat.SLDPro(1., 5., 0.)
        sld = [[layer1, layer2]]
        ddata1 = dataformat.QData(0.05, 0., 0., 0.05 * 0.05)
        ddata2 = dataformat.QData(0.25, 0., 0., 0.05 * 0.25)
        ddata3 = dataformat.QData(0.50, 0., 0., 0.05 * 0.50)
        ddata = [ddata1, ddata2, ddata3]
        a = reflect.Reflect(sld, ddata)
        a.reflect = [data1, data2, data3]
        a.average_ref()
        ddata = []
        k = compare.Compare(ddata, a.averagereflect, 1, 0)
        with self.assertRaises(ValueError) as context:
            k.fit()
        self.assertTrue('No q vectors have been defined -- either read a .dat file or get q vectors.' in str(context.exception))

    def test_fit_noi(self):
        data11 = dataformat.QData(0.05, 0.1, 0., 0.05 * 0.05)
        data12 = dataformat.QData(0.25, 0.05, 0., 0.05 * 0.25)
        data13 = dataformat.QData(0.50, 0.01, 0., 0.05 * 0.50)
        data1 = [data11, data12, data13]
        data21 = dataformat.QData(0.05, 0.12, 0., 0.05 * 0.05)
        data22 = dataformat.QData(0.25, 0.03, 0., 0.05 * 0.25)
        data23 = dataformat.QData(0.50, 0.015, 0., 0.05 * 0.50)
        data2 = [data21, data22, data23]
        data31 = dataformat.QData(0.05, 0.14, 0., 0.05 * 0.05)
        data32 = dataformat.QData(0.25, 0.07, 0., 0.05 * 0.25)
        data33 = dataformat.QData(0.50, 0.005, 0., 0.05 * 0.50)
        data3 = [data31, data32, data33]
        layer1 = dataformat.SLDPro(1., 0., 0.)
        layer2 = dataformat.SLDPro(1., 5., 0.)
        sld = [[layer1, layer2]]
        ddata1 = dataformat.QData(0.05, 0., 0., 0.05 * 0.05)
        ddata2 = dataformat.QData(0.25, 0., 0., 0.05 * 0.25)
        ddata3 = dataformat.QData(0.50, 0., 0., 0.05 * 0.50)
        ddata = [ddata1, ddata2, ddata3]
        a = reflect.Reflect(sld, ddata)
        a.reflect = [data1, data2, data3]
        a.average_ref()
        ddata1 = dataformat.QData(0.05, None, 0., 0.05 * 0.05)
        ddata2 = dataformat.QData(0.25, 0., 0., 0.05 * 0.25)
        ddata3 = dataformat.QData(0.50, 0., 0., 0.05 * 0.50)
        ddata = [ddata1, ddata2, ddata3]
        k = compare.Compare(ddata, a.averagereflect, 1, 0)
        with self.assertRaises(ValueError) as context:
            k.fit()
        self.assertTrue('No experimental data has been set for comparison, please read in a a .dat file.' in str(context.exception))

    def test_return_fitted(self):
        data1 = dataformat.QData(0.05, 3., 0.3, 0.05 * 0.05)
        data2 = dataformat.QData(0.25, 2., 0.2, 0.05 * 0.25)
        data3 = dataformat.QData(0.50, 1., 0.1, 0.05 * 0.50)
        data = [data1, data2, data3]
        sdata1 = dataformat.QData(0.05, 1., 0.1, 0.05 * 0.05)
        sdata2 = dataformat.QData(0.25, 2., 0.2, 0.05 * 0.25)
        sdata3 = dataformat.QData(0.50, 3., 0.3, 0.05 * 0.50)
        sdata = [sdata1, sdata2, sdata3]
        a = compare.Compare(data, sdata, 2., 0.)
        a.return_fitted()
        assert_almost_equal(a.sim_data_fitted[0].i, 2.)
        assert_almost_equal(a.sim_data_fitted[1].i, 4.)
        assert_almost_equal(a.sim_data_fitted[2].i, 6.)
        assert_almost_equal(a.sim_data_fitted[0].di, 0.2)
        assert_almost_equal(a.sim_data_fitted[1].di, 0.4)
        assert_almost_equal(a.sim_data_fitted[2].di, 0.6)

    def test_scale_and_background(self):
        a = np.array([1., 2., 3.])
        scale = 2.
        background = 1.
        b = compare.scale_and_background(a, scale, background)
        assert_almost_equal(b[0], 1.098612289)
        assert_almost_equal(b[1], 1.609437912)
        assert_almost_equal(b[2], 1.945910149)



import numpy as np
from falass import dataformat
import matplotlib.pyplot as plt
from scipy.interpolate import InterpolatedUnivariateSpline


class Reflect:
    """Reflectometry calculation.

    A class for the calculation of reflectometry from the sld profile defined in the falass.sld.SLD class.

    Parameters
    ----------
    sld_profile: array_like falass.dataformat.SLDPro
        An array describing the scattering length density of the simulation cell under study.
    exp_data: array_like falass.dataformat.QData
        An array giving the experimental data from the datfile.
    """
    def __init__(self, sld_profile, exp_data):
        self.sld_profile = sld_profile
        self.exp_data = exp_data
        self.averagereflect = []
        self.reflect = []

    def calc_ref(self):
        """Calculate reflectometry.

        The calculation of the reflectometry profiles based on the sld profiles calculated from each of the timesteps
        under study.
        """
        if len(self.exp_data) > 0:
            self.reflect = []
            prog = 0
            k = 0
            print("Calculating reflectometry\n[ 0 % ]")
            for i in range(0, len(self.sld_profile)):
                k += 1
                prog_new = np.floor(k / (len(self.sld_profile)) * 100)
                if prog_new > prog + 9:
                    prog = prog_new
                    print("[{} {} % ]".format('#' * int(prog / 10), int(prog / 10) * 10))
                refl = convolution(self.exp_data, self.sld_profile[i])
                a = []
                for j in range(0, len(self.exp_data)):
                    a.append(dataformat.QData(self.exp_data[j].q, refl[j], 0, self.exp_data[j].dq))
                self.reflect.append(a)
        else:
            raise ValueError('No q vectors have been defined -- either read a .dat file or get q vectors.')

    def average_ref(self):
        """Average reflectometry profiles.

        The averaging of the reflectometry profiles as calculated by the calc_ref() function.
        """
        if len(self.exp_data) > 0:
            self.averagereflect = []
            for i in range(0, len(self.reflect[0])):
                self.averagereflect.append(dataformat.QData(self.exp_data[i].q, 0, 0, self.exp_data[i].dq))
            for i in range(0, len(self.reflect[0])):
                for j in range(0, len(self.reflect)):
                    self.averagereflect[i].i += self.reflect[j][i].i
                self.averagereflect[i].i /= len(self.reflect)
                for j in range(0, len(self.reflect)):
                    self.averagereflect[i].di += np.square(self.reflect[j][i].i - self.averagereflect[i].i)
                self.averagereflect[i].di = np.sqrt(1. / (len(self.reflect) - 1) * self.averagereflect[i].di)
        else:
            raise ValueError('No q vectors have been defined -- either read a .dat file or get q vectors.')

    def plot_ref(self, rq4=True): #pragma: no cover
        """Plot reflectometry profile.

        The plotting of the calculated reflectometry data without comparison to experimental data.

        Parameters
        ----------
        rq4: bool
            Should the data be transformed to rq4 space.
        """
        if len(self.exp_data) > 0:
            x = []
            y = []
            dy = []
            plt.rc('text')
            plt.rc('font', family='serif')
            if rq4:
                for i in range(0, len(self.exp_data)):
                    x.append(self.averagereflect[i].q)
                    y.append(np.log10(self.averagereflect[i].i * self.averagereflect[i].q ** 4))
                    dy.append((self.averagereflect[i].di * self.averagereflect[i].q ** 4) /
                              (self.averagereflect[i].i * np.log(10)))
            else:
                for i in range(0, len(self.exp_data)):
                    x.append(self.averagereflect[i].q)
                    y.append(np.log10(self.averagereflect[i].i))
                    dy.append(self.averagereflect[i].di / (self.averagereflect[i].i * np.log(10)))
            x = np.asarray(x)
            y = np.asarray(y)
            dy = np.asarray(dy)
            plt.errorbar(x, y, yerr=dy)
            plt.xlabel('$q$ (\AA)')
            plt.ylabel('log($Rq^4$) (\AA$^4$)')
            plt.show()
        else:
            raise ValueError('No q vectors have been defined -- either read a .dat file or get q vectors.')
        return plt

def convolution(exp_data, sld_profile):
    """Convolution/smearing

    The convolution of the reflectometry data by a gaussian of constant width (a percentage of the q-vector)

    Parameters
    ----------
    exp_data: falass.dataformat.QData
        The experimental data from the datfile.
    sld_profile: falass.dataformat.SLDPro
        The SLD profile calculated from the simulation trajectory.
    lt: float
        The thickness of the layers.

    Returns
    -------
    array_like
        The smeared reflectometry profile.
    """
    fwhm = 2 * np.sqrt(2 * np.log(2))

    res = exp_data[0].dq / exp_data[0].q

    if exp_data[0].dq / exp_data[0].q < 0.0005:
        return reflectivity(exp_data, sld_profile)

    gnum = 51
    ggpoint = (gnum - 1) / 2

    def gauss(x, s):
        return 1. / s / np.sqrt(2 * np.pi) * np.exp(-0.5 * x ** 2 / s / s)

    q = []
    for i in range(0, len(exp_data)):
        q.append(exp_data[i].q)

    lowq = np.min(q)
    highq = np.max(q)

    start = np.log10(lowq) - 6 * res / fwhm
    finish = np.log10(highq * (1 + 6 * res / fwhm))
    interp = np.round(np.abs(1 * (np.abs(start-finish)) / (1.7 * res / fwhm / ggpoint)))

    xtemp = np.linspace(start, finish, int(interp))
    xlin = np.power(10., xtemp)

    gaussx = np.linspace(-1.7 * res, 1.7 * res, gnum)
    gaussy = gauss(gaussx, res / fwhm)

    rvals = reflectivity(xlin, sld_profile)
    smeared_rvals = np.convolve(rvals, gaussy, mode='same')
    interpol = InterpolatedUnivariateSpline(xlin, smeared_rvals)

    smeared_output = interpol(q)
    smeared_output *= gaussx[1] - gaussx[0]
    return smeared_output


def reflectivity(exp_data, sld_profile):
    """Abeles optical matrix formalism.

    The calculation of the reflectometry using the Abeles optical matrix method.

    Parameters
    ----------
    exp_data: falass.dataformat.QData
        The experimental data from the datfile.
    sld_profile: falass.dataformat.SLDPro
        The SLD profile calculated from the simulation trajectory.
    lt: float
        The thickness of the layers.

    Returns
    -------
    array_like
        The reflectometry profile.
    """
    layers = np.zeros((len(sld_profile), 4))
    for i in range(0, len(sld_profile)):
        layers[i][0] = sld_profile[i].thick
        layers[i][1] = sld_profile[i].real
        layers[i][2] = sld_profile[i].imag
        layers[i][3] = 0
    qvals = np.asfarray(exp_data).ravel()
    nlayers = len(sld_profile) - 2
    npnts = qvals.size

    kn = make_kn(npnts, nlayers, layers, qvals)

    k = kn[:,0]

    mrtot = [[1, 0], [0, 1]]

    for idx in range(1, nlayers + 2):
        k, mrtot = layer_loop(kn, k, idx, layers, mrtot)

    ref = (mrtot[0][1] * np.conj(mrtot[0][1])) / (mrtot[0][0] * np.conj(mrtot[0][0]))
    return np.real(np.reshape(ref, exp_data.shape))


def layer_loop(kn, k, idx, layers, mrtot):
    """Calculation that is conducted for each layer.

    The is the calculation carried out for each layer in the Abeles optical matrix method calculation.

    Parameters
    ----------
    kn: array_like
        The wavevector at the start of the current layer.
    k: array_like
        The wavevector at the start of the semi-infinite top layer.
    idx: int
        The layer number.
    layers: array_like
        An n by 4 array consisting of information about the layer; thickness, real SLD, imag SLD, and roughness (not
        used in falass), where n is the number of layers.
    mrtot: array_like
        A 2 by 2 array of float comprising the resultant matrix for the layered structure.

    Returns
    -------
    float
        The wavevector at the start of the next layer.
    array_like
        The updated resultant matrix.
    """
    k_next, rj = knext_and_rj(kn, idx, k)

    # work out characteristic matrix of layer
    mi00 = np.exp(k * 1j * np.fabs(layers[idx - 1, 0])) if idx - 1 else 1
    mi11 = np.exp(k * -1j * np.fabs(layers[idx - 1, 0])) if idx - 1 else 1

    mi10 = rj * mi00
    mi01 = rj * mi11

    # matrix multiply mrtot by characteristic matrix
    p0 = mrtot[0][0] * mi00 + mrtot[1][0] * mi01
    p1 = mrtot[0][0] * mi10 + mrtot[1][0] * mi11
    mrtot[0][0] = p0
    mrtot[1][0] = p1

    p0 = mrtot[0][1] * mi00 + mrtot[1][1] * mi01
    p1 = mrtot[0][1] * mi10 + mrtot[1][1] * mi11

    mrtot[0][1] = p0
    mrtot[1][1] = p1

    k = k_next

    return k, mrtot


def make_kn(npnts, nlayers, layers, qvals):
    """Generate the starting kn and sld arrays.

    This will generate the appropriate array for the Abele optical matrix method
    calculation.

    Parameters
    ----------
    npnts: int
        number of q-vectors to be calculated over.
    nlayers: int
        number of layers in system minus two.
    layers: array_like
        An n by 4 array consisting of information about the layer; thickness, real SLD, imag SLD, and roughness (not
        used in falass), where n is the number of layers.
    qvals: array_like
        q-vectors for calculation.

    Returns
    -------
    array_like
        The kn array for the Abeles optical matrix calculation.
    """
    kn = np.zeros((npnts, nlayers + 2), np.complex128)

    sld = np.zeros(nlayers + 2, np.complex128)
    sld[:] += ((layers[:, 1] - layers[0, 1]) + 1j * (layers[:, 2] - layers[0, 2]))

    kn[:] = np.sqrt(qvals[:, np.newaxis] ** 2. / 4. - 4 * np.pi * sld)
    return kn


def knext_and_rj(kn, idx, k):
    """Calculate the k in the next layer and the nature of rj.

    This will determine the wavevector value in the next layer and therefore the value of
    rj which describes the propensity for the wavevector to reflect or refract at a given interface.

    Parameters
    ----------
    kn: array_like
        The wavevector array for a given layer.
    idx: int
        The particular layer.
    k: array_like
        The wavevector for the semi-infinite top layer.

    Returns
    -------
    array_like
        The wavector array in the next layer.
    array_like
        The rj array for the interface.
    """
    k_next = kn[:, idx]
    rj = (k - k_next) / (k + k_next)
    rj *= np.exp(k * k_next)
    return k_next, rj


import numpy as np
from falass import dataformat, readwrite
from scipy.optimize import leastsq

class Analysis:
    """Trajectory analysis

    This is a special class for the analysis of simulation trajectories.

    Parameters
    ----------
    files: falass.readwrite.Files
        Contains information of the files for analysis.
    """
    def __init__(self, files):
        self.files = files
        self.cell = []
        self.atoms = []
        self.number_of_timesteps = 0
        self.times = []
        self.flip = files.flip
        self.number_of_molecules = 0
        self.chain_tilt_angles = []
        self.tail_thickness = []

    def read_full_pdb(self):
        """Parse 3 coordinate pdb.

        Similar to readwrite.read_pdb however instead of only reading a single z-position it reads the 3d position.
        """
        lines = readwrite.line_count(self.files.pdbfile)
        print("Reading PDB file \n")
        file = open(self.files.pdbfile, 'r')
        percentage = 0
        readwrite.print_update(percentage)
        atoms_each_timestep = []
        atoms_each_molecule = []
        for i, line in enumerate(file):
            percentage_new = np.floor(i / lines * 100)
            percentage = readwrite.check_update(percentage, percentage_new)
            if line[0:6] == "ATOM  ":
                atoms_each_molecule.append(get_atom_position(self.cell[self.number_of_timesteps - 1][2], line,
                                                             self.flip))
                if int(line[22:26]) == self.number_of_molecules + 1:
                    if self.number_of_molecules != 0:
                        atoms_each_timestep.append(atoms_each_molecule)
                        atoms_each_molecule = []
                    self.number_of_molecules += 1
            if "TITLE  " in line:
                self.number_of_molecules = 0
                if self.number_of_timesteps == 0:
                    self.number_of_timesteps, new_time = readwrite.iterate_time(self.number_of_timesteps, line)
                    self.times.append(new_time)
                else:
                    self.atoms.append(atoms_each_timestep)
                    atoms_each_timestep = []
                    self.number_of_timesteps, new_time = readwrite.iterate_time(self.number_of_timesteps, line)
                    self.times.append(new_time)
            if "CRYST1  " in line:
                self.cell.append(readwrite.get_cell_parameters(line))
        self.atoms.append(atoms_each_timestep)
        readwrite.print_update(100)
        file.close()
        return

    def get_number_density(self, labels):
        for i in range(0, len(self.atoms)):
            for j in range(0, len(self.atoms[i])):
                for k in range(0, len(labels)):
                    if self.atoms[i][j].atom == labels[k]:
                        print(self.atoms[i][j].atom)

    def get_chain_tilt(self, labels):
        self.chain_tilt_angles = []
        self.tail_thickness = []
        for i in range(0, len(self.atoms)):
            for j in range(0, len(self.atoms[i])):
                for l in range(0, len(labels)):
                    tail_atoms_x = []
                    tail_atoms_y = []
                    tail_atoms_z = []
                    for k in range(0, len(self.atoms[i][j])):
                        if self.atoms[i][j][k].atom in labels[l]:
                            tail_atoms_x.append(np.float(self.atoms[i][j][k].x))
                            tail_atoms_y.append(np.float(self.atoms[i][j][k].y))
                            tail_atoms_z.append(np.float(self.atoms[i][j][k].z))
                    if len(tail_atoms_z) != 0:
                        diffx = np.abs(tail_atoms_x[1] - tail_atoms_x[0])
                        diffy = np.abs(tail_atoms_y[1] - tail_atoms_y[0])
                        diffz = np.abs(tail_atoms_z[1] - tail_atoms_z[0])
                        if diffx > self.cell[i][0]/2:
                            diffx = np.abs(diffx - self.cell[i][0])
                        if diffy > self.cell[i][1]/2:
                            diffy = np.abs(diffy - self.cell[i][1])
                        if diffz > self.cell[i][2]/2:
                            diffz = np.abs(diffz - self.cell[i][2])
                        xy = np.sqrt(np.square(diffx) +
                                     np.square(diffy))
                        theta = np.arctan(xy/diffz)
                        self.tail_thickness.append(diffz)
                        self.chain_tilt_angles.append(np.rad2deg(theta))


def get_atom_position(cell, line, flip):
    """Get the atom position and type

    Reads the text line from the pdb and assigns it to an falass.dataformat.AtomPosition type object.

    Parameters
    ----------
    cell: float
        z-cell dimension.
    line: str
        Line from pdb file.
    flip: bool
        Should the cell be flipped in the z-dimension.

    Returns
    -------
    falass.dataformat.AtomPositions
        Object with atom type and z-position from the given line.
    """
    if flip:
        new_zpos = readwrite.flip_zpos(cell, float(line[46:54]))
        return dataformat.Atom3Positions(line[12:16].strip(), line[30:38].strip(), line[38:46].strip(), new_zpos)
    else:
        return dataformat.Atom3Positions(line[12:16].strip(), line[30:38].strip(), line[38:46].strip(),
                                        float(line[46:54]))